##########################################
##  Kernel Boosting - analysis example  ##
##########################################

rm(list = ls(all = TRUE))

## Set Working directory to source file location
setwd("SUPPLEMENTAL_CODE/")


###########################################
###            'Setup'                  ###
###########################################

## Set seed for reproducibility
set.seed(1104)

## Specify how many cores should be used
CORES <- 1

## Should genetic information be downloaded (download = TRUE) or should example 
## data be loaded (download = FALSE)? 
download <- FALSE    

## Install required packages (if not available on computer)
if (!require("KEGGgraph")) {
    ## try http:// if https:// URLs are not supported
    source("https://bioconductor.org/biocLite.R")
    biocLite("KEGGgraph", suppressUpdates = TRUE)
}
if (!require("biomaRt")) {
    ## try http:// if https:// URLs are not supported
    source("https://bioconductor.org/biocLite.R")
    biocLite("biomaRt", suppressUpdates = TRUE)
}
if (!require("mboost")) {
    install.packages("mboost")
}
if (!require("kangar00")) {
    install.packages("kangar00")    
}

## Load mboost (> 2.7-0)
library("mboost")

## Load kangar00
library("kangar00")
    
############################################
###        Prepare study data            ###
############################################

if (download == TRUE){

    ## Download example pathways from KEGG and create list including all 
    ## pathways to be considered in the analysis      
    list.of.pathways <- list( )   
    for(pw in c("hsa04662", "hsa04917", "hsa05145", "hsa05210", "hsa04024")){  
        new.pathway <- get_network_matrix(pathway(id = paste0(pw), 
                                                  adj = matrix(0),
                                                  sign = as.vector(matrix(0)[matrix(0) != 0])), 
                                          directed = FALSE)
        list.of.pathways <- c(list.of.pathways, new.pathway)    
    }
    
    ## Download information on pathways needed for marker annotation  
    pathway.info.df <- data.frame()
    for(i in 1:length(list.of.pathways)){  
        new.pathway.info <- pathway_info(list.of.pathways[[i]]@id)
        pathway.info.df  <- rbind(pathway.info.df, new.pathway.info@info)
    }
    pathway.info <- new('pathway_info', info = pathway.info.df)
    
    ## Download marker positions needed for annotation from Ensemble database 
    snps <- get(load("snps.Rda")) # load SNP names (rs-numbers)
    snp.info <- snp_info(snps)
    
} else {  
    
    ## Load example pathways from KEGG to be considered in analysis   
    list.of.pathways <- get(load("list.of.pathways.Rda"))  
    
    ## Load pathway information file   
    pathway.info <- get(load("pathway.info.Rda"))
    
    ## Load marker positions needed for annotation 
    snp.info <- get(load("snp.info.Rda"))            
}  


############################################
###       Combine genetic data           ###
############################################
    
## Annotate SNPs to pathways according to base-pair positions
anno <- get_anno(snp.info, pathway.info)   

## Load (simulated) GWAS data ["Null case", i.e., no association with phenotype]
geno  <- get(load("genotype.data.Rda")) 
pheno <- get(load("phenotype.data.Rda"))                           

## Combine study information to GWASdata object               
gwd  <- GWASdata(pheno = pheno, anno = anno, geno = geno, desc = "study_data")  
data <- list(case_control_status = pheno$case_control_status, gwd = gwd)
  

############################################
###        Apply Kernel Boosting         ###
############################################ 

## Create formula for kernel boosting model
kernels <- paste0('bkernel(gwd, kernel = "net", pathway = list.of.pathways[[',
                  seq(1:length(list.of.pathways)), "]], df = 4)")
kernels <- paste(kernels, collapse = " + ")
(fm <- as.formula(paste("case_control_status ~ ", kernels)))
data$case_control_status <- factor(data$case_control_status)

## Fit offset model for environmental variables
logit <- glm(case_control_status ~ age + sex, data = gwd@pheno, 
             family = binomial())
summary(logit)

## Initialize boosting model with:
## - formula fm for the kernel base-lerners as given above
## - GWAS data set
## - run 0 boosting iterations, i.e., just initialize the model
##   and show trace so that one sees the progress    
## - offset (predictions from offset model as fitted above)
## - logistic regression model (Binomial)
##
## Note that the initial set-up takes a while
##
mod <- mboost(fm, data = data, control = boost_control(mstop = 0, trace = TRUE),
              offset = predict(logit), family = Binomial())
           

############################################
###         Run cross-validation         ###
############################################

## Tune model via cross-validation
cat("Cross-validation started.\n")
## For computational reasons we run 5-fold subsamling. To use more folds increase B.
folds_subsampl <- cv(model.weights(mod), type = "subsampling", B = 5) 
## Of note: Usually grid must be larger than 0:20 (these are the tested mstop values)
cvr <- cvrisk(mod, folds = folds_subsampl, grid = 0:20, mc.cores = CORES)
plot(cvr)


############################################
###       Kernel Boosting Results        ###
############################################  

## Set model to optimal mstop 
cat("Set m_stop to ", mstop(cvr), ".\n", sep = "")
mstop(mod) <- mstop(cvr)  # (in practive this is usually bigger than 0)

## Extract coefficient estimates of selected base-learners
cf <- coef(mod)
## Extract selected base-learners
(sel <- selected(mod))

## Predicted class probability (p)
predicted_prob  <- predict(mod, type = "response")
## Predicted class (with cutoff p = 1/2)
predicted_status <- predict(mod, type = "class")

# Of note: With the simulated null case, no base-learners are selected and
#          hence, the model consists only of the environmental offset model.
#          Usually, we could interpret which base-learners were selected,
#          how often, and when. Furthermore, we could assess the norm of the 
#          individual coefficients (separately for each base-learner) in order
#          to gain information on the "effect size" and importance of pathways.
  
  
############################################  
###        Compare to LKMT results       ###
############################################  

cat("Start single pathway tests using LKMT.\n")
results <- list()

for(i in 1:length(list.of.pathways)){
    K.net <- calc_kernel(gwd, list.of.pathways[[i]], type = 'net', 
                         calculation = 'cpu')
    p.net <- lkmt_test (case_control_status ~ age + sex, K.net, 
                        gwd, method = 'satt')
    
    results <- c(results, p.net)                
}
cat("Following p-values were calculated:\n")
print(results) 
